""" LSF mn adaptor implementation
"""

import re
import os
import time
import threading
# noinspection PyDeprecation
from cgi import parse_qs

# noinspection PyPackageRequirements
import radical.utils.threads as sut
import radical.saga as saga
import saga.url as surl
import saga.utils.pty_shell
import saga.adaptors.base
import saga.adaptors.cpi.job

import saga.adaptors.cpi.decorators

SYNC_CALL = saga.adaptors.cpi.decorators.SYNC_CALL
ASYNC_CALL = saga.adaptors.cpi.decorators.ASYNC_CALL

SYNC_WAIT_UPDATE_INTERVAL = 1  # seconds
MONITOR_UPDATE_INTERVAL = 3  # seconds


# --------------------------------------------------------------------
#
# noinspection PyProtectedMember,PyPep8Naming,PyMissingOrEmptyDocstring
class _job_state_monitor(threading.Thread):
    """ thread that periodically monitors job states
    """

    def __init__(self, job_service):

        self.logger = job_service._logger
        self.js = job_service
        self._stop = sut.Event()

        super(_job_state_monitor, self).__init__()
        self.setDaemon(True)

    def stop(self):
        self._stop.set()

    def stopped(self):
        return self._stop.isSet()

    def run(self):
        while self.stopped() is False:
            try:
                # do bulk updates here! we don't want to pull information
                # job by job. that would be too inefficient!
                jobs = self.js.jobs
                job_keys = jobs.keys()

                for job in job_keys:
                    # if the job hasn't been started, we can't update its
                    # state. we can tell if a job has been started if it
                    # has a job id
                    if jobs[job].get('job_id', None) is not None:
                        # we only need to monitor jobs that are not in a
                        # terminal state, so we can skip the ones that are
                        # either done, failed or canceled
                        state = jobs[job]['state']
                        if state not in [saga.job.DONE, saga.job.FAILED, saga.job.CANCELED]:

                            job_info = self.js._job_get_info(job)
                            self.logger.info(
                                "Job monitoring thread updating Job %s (state: %s)" % (job, job_info['state']))

                            if job_info['state'] != state:
                                # fire job state callback if 'state' has changed
                                job._api()._attributes_i_set('state', job_info['state'], job._api()._UP, True)

                            # update job info
                            self.js.jobs[job] = job_info

                time.sleep(MONITOR_UPDATE_INTERVAL)
            except Exception as e:
                self.logger.warning("Exception caught in job monitoring thread: %s" % e)


# --------------------------------------------------------------------
#
def log_error_and_raise(message, exception, logger):
    """
    logs an 'error' message and subsequently throws an exception

    :param message: message to log
    :param exception: exception to rise
    :param logger: logger to use
    """
    logger.error(message)
    raise exception(message)


# --------------------------------------------------------------------
#
def _mn_to_saga_jobstate(mnjs):
    """ translates a mn one-letter state to saga
    """
    if mnjs in ['RUN']:
        return saga.job.RUNNING
    elif mnjs in ['WAIT', 'PEND']:
        return saga.job.PENDING
    elif mnjs in ['DONE']:
        return saga.job.DONE
    elif mnjs in ['ZOMBI', 'EXIT']:
        return saga.job.FAILED
    elif mnjs in ['USUSP', 'SSUSP', 'PSUSP']:
        return saga.job.SUSPENDED
    else:
        return saga.job.UNKNOWN


# --------------------------------------------------------------------
#
def _mnscript_generator(jd, queue=None):
    """ generates an LSF script from a SAGA job description
    """
    mn_params = str()
    exec_n_args = str()

    if jd.executable is not None:
        exec_n_args += "%s " % jd.executable
    if jd.arguments is not None:
        for arg in jd.arguments:
            exec_n_args += "%s " % arg

    if jd.name is not None:
        mn_params += "#BSUB -J %s \n" % jd.name

    if jd.environment is not None:
        env_variable_list = "export "
        for key in jd.environment.keys():
            env_variable_list += " %s=%s " % (key, jd.environment[key])
    else:
        env_variable_list = ""

    # a workaround is to do an explicit 'cd'
    if jd.working_directory is not None:
        mn_params += "#BSUB -cwd %s \n" % jd.working_directory

    if jd.output is not None:
        # if working directory is set, we want stdout to end up in
        # the working directory as well, unless it containes a specific
        # path name.
        if jd.working_directory is not None:
            if os.path.isabs(jd.output):
                mn_params += "#BSUB -o %s \n" % jd.output
            else:
                # user provided a relative path for STDOUT. in this case
                # we prepend the working directory path before passing
                # it on to LSF.
                mn_params += "#BSUB -o %s/%s \n" % (jd.working_directory, jd.output)
        else:
            mn_params += "#BSUB -o %s \n" % jd.output

    if jd.error is not None:
        # if working directory is set, we want stderr to end up in
        # the working directory as well, unless it contains a specific
        # path name.
        if jd.working_directory is not None:
            if os.path.isabs(jd.error):
                mn_params += "#BSUB -e %s \n" % jd.error
            else:
                # user provided a relative path for STDERR. in this case
                # we prepend the working directory path before passing
                # it on to LSF.
                mn_params += "#BSUB -e %s/%s \n" % (jd.working_directory, jd.error)
        else:
            mn_params += "#BSUB -e %s \n" % jd.error

    if jd.wall_time_limit is not None:
        hours = jd.wall_time_limit / 60
        minutes = jd.wall_time_limit % 60
        mn_params += "#BSUB -W %s:%s \n" \
                     % (str(hours), str(minutes))

    if (jd.queue is not None) and (queue is not None):
        mn_params += "#BSUB -q %s \n" % queue
    elif (jd.queue is not None) and (queue is None):
        mn_params += "#BSUB -q %s \n" % jd.queue
    elif (jd.queue is None) and (queue is not None):
        mn_params += "#BSUB -q %s \n" % queue

    if jd.project is not None:
        if ':' not in jd.project:
            account = jd.project
        else:
            account, reservation, exclusivity = jd.project.split(':')
            if reservation != '':
                mn_params += "#BSUB -U %s \n" % str(reservation)
            if exclusivity == 'true':
                mn_params += "#BSUB -x \n"

        mn_params += "#BSUB -P %s \n" % str(account)
    if jd.job_contact is not None:
        mn_params += "#BSUB -u %s \n" % str(jd.job_contact)

    # if total_cpu_count is not defined, we assume 1
    if jd.total_cpu_count is None:
        jd.total_cpu_count = 1

    mn_params += "#BSUB -n %s \n" % str(jd.total_cpu_count)

    if jd.processes_per_host:
        mn_params += '#BSUB -R "span[ptile=%s]"\n' % str(jd.processes_per_host)

    # escape all double quotes and dollarsigns, otherwise 'echo |'
    # further down won't work
    # only escape '$' in args and exe. not in the params
    # exec_n_args = workdir_directives exec_n_args
    exec_n_args = exec_n_args.replace('$', '\\$')

    mnscript = "\n#!/bin/bash \n%s\n%s\n%s" % (mn_params, env_variable_list, exec_n_args)

    mnscript = mnscript.replace('"', '\\"')
    return mnscript

# --------------------------------------------------------------------
# some private defs
#
_PTY_TIMEOUT = 2.0

# --------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME = "autosubmit.platforms.mn_adaptor"
_ADAPTOR_SCHEMAS = ["mn", "mn+ssh", "mn+gsissh"]
_ADAPTOR_OPTIONS = []

# --------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES = {
    "jdes_attributes": [saga.job.NAME,
                        saga.job.EXECUTABLE,
                        saga.job.ARGUMENTS,
                        saga.job.ENVIRONMENT,
                        saga.job.INPUT,
                        saga.job.OUTPUT,
                        saga.job.ERROR,
                        saga.job.QUEUE,
                        saga.job.PROJECT,
                        saga.job.WALL_TIME_LIMIT,
                        saga.job.WORKING_DIRECTORY,
                        saga.job.SPMD_VARIATION,  # TODO: 'hot'-fix for BigJob
                        saga.job.PROCESSES_PER_HOST,
                        saga.job.TOTAL_CPU_COUNT],
    "job_attributes": [saga.job.EXIT_CODE,
                       saga.job.EXECUTION_HOSTS,
                       saga.job.CREATED,
                       saga.job.STARTED,
                       saga.job.FINISHED],
    "metrics": [saga.job.STATE],
    "callbacks": [saga.job.STATE],
    "contexts": {"ssh": "SSH public/private keypair",
                 "x509": "GSISSH X509 proxy context",
                 "userpass": "username/password pair (ssh)"}
}

# --------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC = {
    "name": _ADAPTOR_NAME,
    "cfg_options": _ADAPTOR_OPTIONS,
    "capabilities": _ADAPTOR_CAPABILITIES,
    "description": """
The MN adaptor allows to run and manage jobs on MareNostrum 3
""",
    "schemas": {"mn": "connect to a local cluster",
                "mn+ssh": "connect to a remote cluster via SSH",
                "mn+gsissh": "connect to a remote cluster via GSISSH"}
}

# --------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA
#
_ADAPTOR_INFO = {
    "name": _ADAPTOR_NAME,
    "version": "v0.2",
    "schemas": _ADAPTOR_SCHEMAS,
    "capabilities": _ADAPTOR_CAPABILITIES,
    "cpis": [
        {
            "type": "saga.job.Service",
            "class": "MNJobService"
        },
        {
            "type": "saga.job.Job",
            "class": "MNJob"
        }
    ]
}


###############################################################################
# The adaptor class
# noinspection PyMissingOrEmptyDocstring
class Adaptor(saga.adaptors.base.Base):
    """ this is the actual adaptor class, which gets loaded by SAGA (i.e. by
        the SAGA engine), and which registers the CPI implementation classes
        which provide the adaptor's functionality.
    """

    # ----------------------------------------------------------------
    #
    def __init__(self):
        # noinspection PyCallByClass,PyTypeChecker
        saga.adaptors.base.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        self.id_re = re.compile('^\[(.*)\]-\[(.*?)\]$')
        self.opts = self.get_config(_ADAPTOR_NAME)

    # ----------------------------------------------------------------
    #
    def sanity_check(self):
        # FIXME: also check for gsissh
        pass

    # ----------------------------------------------------------------
    #
    def parse_id(self, job_id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(job_id)

        if not match or len(match.groups()) != 2:
            raise saga.BadParameter("Cannot parse job id '%s'" % job_id)

        return match.group(1), match.group(2)


###############################################################################
#
# noinspection PyMethodOverriding,PyMethodOverriding,PyProtectedMember, PyMissingOrEmptyDocstring
class MNJobService(saga.adaptors.cpi.job.Service):
    """ implements saga.adaptors.cpi.job.Service
    """

    # ----------------------------------------------------------------
    #
    # noinspection PyMissingConstructor
    def __init__(self, api, adaptor):

        self._mt = None
        _cpi_base = super(MNJobService, self)
        _cpi_base.__init__(api, adaptor)

        self._adaptor = adaptor

    # ----------------------------------------------------------------
    #
    def __del__(self):

        self.close()

    # ----------------------------------------------------------------
    #
    def close(self):

        if self.mt:
            self.mt.stop()
            self.mt.join(10)  # don't block forever on join()

        self._logger.info("Job monitoring thread stopped.")

        self.finalize(True)

    # ----------------------------------------------------------------
    #
    def finalize(self, kill_shell=False):

        if kill_shell:
            if self.shell:
                self.shell.finalize(True)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        """
        service instance constructor

        :param adaptor_state: state of the adaptor
        :param rm_url: machine url
        :param session: SAGA session to use
        """
        self.rm = rm_url
        self.session = session
        self.ppn = 1
        self.queue = None
        self.shell = None
        self.jobs = dict()

        # the monitoring thread - one per service instance
        # noinspection PyTypeChecker
        self.mt = _job_state_monitor(job_service=self)
        self.mt.start()

        rm_scheme = rm_url.scheme
        pty_url = surl.Url(rm_url)

        # this adaptor supports options that can be passed via the
        # 'query' component of the job service URL.
        if rm_url.query is not None:
            # noinspection PyDeprecation
            for key, val in parse_qs(rm_url.query).iteritems():
                if key == 'queue':
                    self.queue = val[0]
                elif key == 'span':
                    self.span = val[0]

        # we need to extrac the scheme for PTYShell. That's basically the
        # job.Serivce Url withou the mn+ part. We use the PTYShell to execute
        # mn commands either locally or via gsissh or ssh.
        if rm_scheme == "mn":
            pty_url.scheme = "fork"
        elif rm_scheme == "mn+ssh":
            pty_url.scheme = "ssh"
        elif rm_scheme == "mn+gsissh":
            pty_url.scheme = "gsissh"

        # these are the commands that we need in order to interact with LSF.
        # the adaptor will try to find them during initialize(self) and bail
        # out in case they are note avaialbe.
        self._commands = {'bqueues': '9.1.2.0',
                          'bjobs': '9.1.2.0',
                          'bsub': '9.1.2.0',
                          'bkill': '9.1.2.0'}

        self.shell = saga.utils.pty_shell.PTYShell(pty_url, self.session)

        # self.shell.set_initialize_hook(self.initialize)
        # self.shell.set_finalize_hook(self.finalize)

        self.initialize()
        return self.get_api()

    # ----------------------------------------------------------------
    #
    def initialize(self):
        # check if all required mn tools are available
        # for cmd in self._commands.keys():
        #     ret, out, _ = self.shell.run_sync("which %s " % cmd)
        #     if ret != 0:
        #         message = "Couldn't find LSF tools: %s" % out
        #         log_error_and_raise(message, saga.NoSuccess, self._logger)
        #     else:
        #         path = out.strip()  # strip removes newline
        #     ret, out, _ = self.shell.run_sync("%s -V" % cmd)
        #     if ret != 0:
        #         message = "Couldn't find LSF tools: %s" % out
        #         log_error_and_raise(message, saga.NoSuccess, self._logger)
        #     else:
        #         # version is reported as: "version: x.y.z"
        #         version = out.split("\n")[0]
        #
        #         # add path and version to the command dictionary
        #         self._commands[cmd] = {"path":    cmd,
        #                                "version": version}

        self._logger.info("Found LSF tools: %s" % self._commands)

        # see if we can get some information about the cluster, e.g.,
        # different queues, number of processes per node, etc.
        # TODO: this is quite a hack. however, it *seems* to work quite
        #       well in practice.
        # ret, out, _ = self.shell.run_sync('unset GREP_OPTIONS; %s -a | grep -E "(np|pcpu)"' % \
        #    self._commands['pbsnodes']['path'])
        # if ret != 0:
        #
        #    message = "Error running pbsnodes: %s" % out
        #    log_error_and_raise(message, saga.NoSuccess, self._logger)
        # else:
        # this is black magic. we just assume that the highest occurence
        # of a specific np is the number of processors (cores) per compute
        # node. this equals max "PPN" for job scripts
        #    ppn_list = dict()
        #    for line in out.split('\n'):
        #        np = line.split(' = ')
        #        if len(np) == 2:
        #            np = np[1].strip()
        #            if np in ppn_list:
        #                ppn_list[np] += 1
        #            else:
        #                ppn_list[np] = 1
        #    self.ppn = max(ppn_list, key=ppn_list.get)
        #    self._logger.debug("Found the following 'ppn' configurations: %s. \

    # Using %s as default ppn."
    #           % (ppn_list, self.ppn))

    # ----------------------------------------------------------------
    #
    def _job_run(self, job_obj):
        """ runs a job via qsub
        """
        # get the job description
        jd = job_obj.jd

        # normalize working directory path
        if jd.working_directory:
            jd.working_directory = os.path.normpath(jd.working_directory)

        if (self.queue is not None) and (jd.queue is not None):
            self._logger.warning("Job service was instantiated explicitly with \
'queue=%s', but job description tries to a differnt queue: '%s'. Using '%s'." %
                                 (self.queue, jd.queue, self.queue))

        try:
            # create an LSF job script from SAGA job description
            script = _mnscript_generator(jd=jd, queue=self.queue)

            self._logger.info("Generated LSF script: %s" % script)
        except Exception as ex:
            script = ''
            log_error_and_raise(str(ex), saga.BadParameter, self._logger)

        # try to create the working directory (if defined)
        # WARNING: this assumes a shared filesystem between login node and
        #          compute nodes.
        if jd.working_directory is not None:
            self._logger.info("Creating working directory %s" % jd.working_directory)
            ret, out, _ = self.shell.run_sync("mkdir -p %s" % jd.working_directory)
            if ret != 0:
                # something went wrong
                message = "Couldn't create working directory - %s" % out
                log_error_and_raise(message, saga.NoSuccess, self._logger)

        # Now we want to execute the script. This process consists of two steps:
        # (1) we create a temporary file with 'mktemp' and write the contents of
        #     the generated PBS script into it
        # (2) we call 'qsub <tmpfile>' to submit the script to the queueing system
        # noinspection PyPep8
        cmdline = """SCRIPTFILE=`mktemp -t SAGA-Python-LSFJobScript.XXXXXX` && echo "%s" > $SCRIPTFILE && %s < $SCRIPTFILE && rm -f $SCRIPTFILE""" % (script, 'bsub')
        ret, out, _ = self.shell.run_sync(cmdline)

        if ret != 0:
            # something went wrong
            message = "Error running job via 'bsub': %s. Commandline was: %s" \
                      % (out, cmdline)
            log_error_and_raise(message, saga.NoSuccess, self._logger)
        else:
            # parse the job id. bsub's output looks like this:
            # Job <901545> is submitted to queue <regular>
            lines = out.split("\n")
            lines = filter(lambda l: l != '', lines)  # remove empty

            self._logger.info('bsub: %s' % ''.join(lines))

            mn_job_id = None
            for line in lines:
                if re.search('Job <.+> is submitted to', line):
                    mn_job_id = re.findall(r'<(.*?)>', line)[0]
                    break

            if not mn_job_id:
                raise Exception("Failed to detect job id after submission.")

            job_id = "[%s]-[%s]" % (self.rm, mn_job_id)

            self._logger.info("Submitted LSF job with id: %s" % job_id)

            # update job dictionary
            self.jobs[job_obj]['job_id'] = job_id
            self.jobs[job_obj]['submitted'] = job_id

            # set status to 'pending' and manually trigger callback
            # self.jobs[job_obj]['state'] = saga.job.PENDING
            # job_obj._api()._attributes_i_set('state', self.jobs[job_obj]['state'], job_obj._api()._UP, True)

            # return the job id
            return job_id

    # ----------------------------------------------------------------
    #
    def _retrieve_job(self, job_id):
        """ see if we can get some info about a job that we don't
            know anything about
        """
        rm, pid = self._adaptor.parse_id(job_id)

        ret, out, _ = self.shell.run_sync(
            "%s -noheader %s" % ('bjobs', pid))

        if ret != 0:
            message = "Couldn't reconnect to job '%s': %s" % (job_id, out)
            log_error_and_raise(message, saga.NoSuccess, self._logger)

        else:
            # the job seems to exist on the backend. let's gather some data
            job_info = {
                'state': saga.job.UNKNOWN,
                'exec_hosts': None,
                'returncode': None,
                'create_time': None,
                'start_time': None,
                'end_time': None,
                'gone': False
            }

            results = out.split()
            job_info['state'] = _mn_to_saga_jobstate(results[2])
            job_info['exec_hosts'] = results[5]
            job_info['create_time'] = results[7]

            return job_info

    # ----------------------------------------------------------------
    #
    def _job_get_info(self, job_obj):
        """ get job attributes via bjob
        """

        # if we don't have the job in our dictionary, we don't want it
        if job_obj not in self.jobs:
            message = "Unknown job object: %s. Can't update state." % job_obj._id
            log_error_and_raise(message, saga.NoSuccess, self._logger)

        # prev. info contains the info collect when _job_get_info
        # was called the last time
        prev_info = self.jobs[job_obj]

        # if the 'gone' flag is set, there's no need to query the job
        # state again. it's gone forever
        if prev_info['gone'] is True:
            return prev_info

        # curr. info will contain the new job info collect. it starts off
        # as a copy of prev_info (don't use deepcopy because there is an API
        # object in the dict -> recursion)
        curr_info = dict()
        curr_info['job_id'] = prev_info.get('job_id')
        curr_info['state'] = prev_info.get('state')
        curr_info['exec_hosts'] = prev_info.get('exec_hosts')
        curr_info['returncode'] = prev_info.get('returncode')
        curr_info['create_time'] = prev_info.get('create_time')
        curr_info['start_time'] = prev_info.get('start_time')
        curr_info['end_time'] = prev_info.get('end_time')
        curr_info['gone'] = prev_info.get('gone')

        rm, pid = self._adaptor.parse_id(job_obj._id)

        # run the LSF 'bjobs' command to get some infos about our job
        # the result of bjobs <id> looks like this:
        #
        # JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME
        # 901545  oweidne DONE  regular    yslogin5-ib ys3833-ib   *FILENAME  Nov 11 12:06
        #
        # If we add the -nodeader flag, the first row is ommited

        ret, out, _ = self.shell.run_sync("%s -noheader %s" % ('bjobs', pid))

        if ret != 0:
            if "Illegal job ID" in out or "is not found" in out:
                # Let's see if the previous job state was running or pending. in
                # that case, the job is gone now, which can either mean DONE,
                # or FAILED. the only thing we can do is set it to 'DONE'
                curr_info['gone'] = True
                # we can also set the end time
                self._logger.warning("Previously running job has disappeared." +
                                     "This probably means that the backend doesn't " +
                                     "store informations about finished jobs." +
                                     "Setting state to 'DONE'.")

                if prev_info['state'] in [saga.job.RUNNING, saga.job.PENDING]:
                    curr_info['state'] = saga.job.DONE
                else:
                    curr_info['state'] = saga.job.FAILED
            else:
                # something went wrong
                message = "Error retrieving job info via 'bjobs': %s" % out
                log_error_and_raise(message, saga.NoSuccess, self._logger)
        else:
            # parse the result
            results = out.split()
            curr_info['state'] = _mn_to_saga_jobstate(results[2])
            curr_info['exec_hosts'] = results[5]

        # return the new job info dict
        return curr_info

    # ----------------------------------------------------------------
    #
    def _job_get_state(self, job_obj):
        """ get the job's state
        """
        return self.jobs[job_obj]['state']

    # ----------------------------------------------------------------
    #
    def _job_get_exit_code(self, job_obj):
        """ get the job's exit code
        """
        ret = self.jobs[job_obj]['returncode']

        # FIXME: 'None' should cause an exception
        if ret is None:
            return None
        else:
            return int(ret)

    # ----------------------------------------------------------------
    #
    def _job_get_execution_hosts(self, job_obj):
        """ get the job's exit code
        """
        return self.jobs[job_obj]['exec_hosts']

    # ----------------------------------------------------------------
    #
    def _job_get_create_time(self, job_obj):
        """ get the job's creation time
        """
        return self.jobs[job_obj]['create_time']

    # ----------------------------------------------------------------
    #
    def _job_get_start_time(self, job_obj):
        """ get the job's start time
        """
        return self.jobs[job_obj]['start_time']

    # ----------------------------------------------------------------
    #
    def _job_get_end_time(self, job_obj):
        """ get the job's end time
        """
        return self.jobs[job_obj]['end_time']

    # ----------------------------------------------------------------
    #
    def _job_cancel(self, job_obj):
        """ cancel the job via 'qdel'
        """
        rm, pid = self._adaptor.parse_id(job_obj._id)

        ret, out, _ = self.shell.run_sync("%s %s\n" % ('qdel', pid))

        if ret != 0:
            message = "Error canceling job via 'qdel': %s" % out
            log_error_and_raise(message, saga.NoSuccess, self._logger)

        # assume the job was succesfully canceled
        self.jobs[job_obj]['state'] = saga.job.CANCELED

    # ----------------------------------------------------------------
    #
    def _job_wait(self, job_obj, timeout):
        """ wait for the job to finish or fail
        """
        time_start = time.time()
        time_now = time_start
        self._adaptor.parse_id(job_obj._id)

        while True:
            # state = self._job_get_state(job_id=job_id, job_obj=job_obj)
            state = self.jobs[job_obj]['state']  # this gets updated in the bg.

            if state == saga.job.DONE or \
               state == saga.job.FAILED or \
               state == saga.job.CANCELED:
                return True

            # avoid busy poll
            time.sleep(SYNC_WAIT_UPDATE_INTERVAL)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        """
        implements saga.adaptors.cpi.job.Service.get_url()

        :param jd: job description
        """
        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service": self,
                         "job_description": jd,
                         "job_schema": self.rm.schema,
                         "reconnect": False
                         }

        # create a new job object
        job_obj = saga.job.Job(_adaptor=self._adaptor,
                               _adaptor_state=adaptor_state)

        # add job to internal list of known jobs.
        self.jobs[job_obj._adaptor] = {
            'state': saga.job.NEW,
            'job_id': None,
            'exec_hosts': None,
            'returncode': None,
            'create_time': None,
            'start_time': None,
            'end_time': None,
            'gone': False,
            'submitted': False
        }

        return job_obj

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, jobid):
        """
        Implements saga.adaptors.cpi.job.Service.get_job()

        :param jobid: job identifier
        :type jobid: str
        """

        # try to get some information about this job
        job_info = self._retrieve_job(jobid)

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service": self,
                         # TODO: fill job description
                         "job_description": saga.job.Description(),
                         "job_schema": self.rm.schema,
                         "reconnect": True,
                         "reconnect_jobid": jobid
                         }

        job = saga.job.Job(_adaptor=self._adaptor,
                           _adaptor_state=adaptor_state)

        # throw it into our job dictionary.
        self.jobs[job._adaptor] = job_info
        return job

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        """ implements saga.adaptors.cpi.job.Service.get_url()
        """
        return self.rm

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        """ implements saga.adaptors.cpi.job.Service.list()
        """
        ids = []

        ret, out, _ = self.shell.run_sync("%s -a -noheader" % 'bjobs')

        if ret != 0 and len(out) > 0:
            message = "failed to list jobs via 'bjobs': %s" % out
            log_error_and_raise(message, saga.NoSuccess, self._logger)
        elif ret != 0 and len(out) == 0:

            pass
        else:
            for line in out.split("\n"):
                # output looks like this:
                # 112059.svc.uc.futuregrid testjob oweidner 0 Q batch
                # 112061.svc.uc.futuregrid testjob oweidner 0 Q batch
                if len(line.split()) > 1:
                    jobid = "[%s]-[%s]" % (self.rm, line.split()[0])
                    ids.append(str(jobid))

        return ids

        # # ----------------------------------------------------------------
        # #
        # def container_run (self, jobs) :
        #     self._logger.debug ("container run: %s"  %  str(jobs))
        #     # TODO: this is not optimized yet
        #     for job in jobs:
        #         job.run ()
        #
        #
        # # ----------------------------------------------------------------
        # #
        # def container_wait (self, jobs, mode, timeout) :
        #     self._logger.debug ("container wait: %s"  %  str(jobs))
        #     # TODO: this is not optimized yet
        #     for job in jobs:
        #         job.wait ()
        #
        #
        # # ----------------------------------------------------------------
        # #
        # def container_cancel (self, jobs) :
        #     self._logger.debug ("container cancel: %s"  %  str(jobs))
        #     raise saga.NoSuccess ("Not Implemented");


###############################################################################
#
# noinspection PyMethodOverriding,PyProtectedMember,PyMissingOrEmptyDocstring
class MNJob(saga.adaptors.cpi.job.Job):
    """ implements saga.adaptors.cpi.job.Job
    """

    # noinspection PyMissingConstructor
    def __init__(self, api, adaptor):

        # initialize parent class
        _cpi_base = super(MNJob, self)
        _cpi_base.__init__(api, adaptor)

    def _get_impl(self):
        return self

    @SYNC_CALL
    def init_instance(self, job_info):
        """
        implements saga.adaptors.cpi.job.Job.init_instance()

        :param job_info: job descriptiom
        :type job_info: dict
        """
        # init_instance is called for every new saga.job.Job object
        # that is created
        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        if job_info['reconnect'] is True:
            self._id = job_info['reconnect_jobid']
            self._name = self.jd.get(saga.job.NAME)
            self._started = True
        else:
            self._id = None
            self._name = self.jd.get(saga.job.NAME)
            self._started = False

        return self.get_api()

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        """ implements saga.adaptors.cpi.job.Job.get_state()
        """
        return self.js._job_get_state(job_obj=self)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):
        """
        implements saga.adaptors.cpi.job.Job.wait()

        :param timeout: time to wait
        :type timeout: int
        """
        if self._started is False:
            log_error_and_raise("Can't wait for job that hasn't been started",
                                saga.IncorrectState, self._logger)
        else:
            self.js._job_wait(job_obj=self, timeout=timeout)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):
        """
        implements saga.adaptors.cpi.job.Job.cancel()

        :param timeout: time to wait
        :type timeout: int
        """
        if self._started is False:
            log_error_and_raise("Can't wait for job that hasn't been started",
                                saga.IncorrectState, self._logger)
        else:
            self.js._job_cancel(self)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):
        """ implements saga.adaptors.cpi.job.Job.run()
        """
        self._id = self.js._job_run(self)
        self._started = True

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):
        """ implements saga.adaptors.cpi.job.Job.get_service_url()
        """
        return self.js.rm

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):
        """ implements saga.adaptors.cpi.job.Job.get_id()
        """
        return self._id

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name (self):
        """ Implements saga.adaptors.cpi.job.Job.get_name() """
        return self._name

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):
        """ implements saga.adaptors.cpi.job.Job.get_exit_code()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_exit_code(self)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):
        """ implements saga.adaptors.cpi.job.Job.get_created()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_create_time(self)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):
        """ implements saga.adaptors.cpi.job.Job.get_started()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_start_time(self)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):
        """ implements saga.adaptors.cpi.job.Job.get_finished()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_end_time(self)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):
        """ implements saga.adaptors.cpi.job.Job.get_execution_hosts()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_execution_hosts(self)



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import os

from xml.dom.minidom import parseString

from autosubmitAPIwu.platforms.paramiko_platform import ParamikoPlatform
from autosubmitAPIwu.platforms.headers.slurm_header import SlurmHeader
from autosubmitAPIwu.platforms.wrappers.wrapper_factory import SlurmWrapperFactory
from autosubmitAPIwu.config.basicConfig import BasicConfig


class SlurmPlatform(ParamikoPlatform):
    """
    Class to manage jobs to host using SLURM scheduler

    :param expid: experiment's identifier
    :type expid: str
    """

    def __init__(self, expid, name, config):
        ParamikoPlatform.__init__(self, expid, name, config)
        self._header = SlurmHeader()
        self._wrapper = SlurmWrapperFactory(self)
        self.job_status = dict()
        self.job_status['COMPLETED'] = ['COMPLETED']
        self.job_status['RUNNING'] = ['RUNNING']
        self.job_status['QUEUING'] = ['PENDING', 'CONFIGURING', 'RESIZING']
        self.job_status['FAILED'] = ['FAILED', 'CANCELLED', 'NODE_FAIL', 'PREEMPTED',
                                     'SUSPENDED', 'TIMEOUT', 'OUT_OF_MEMORY', 'OUT_OF_ME+', 'OUT_OF_ME']
        self._pathdir = "\$HOME/LOG_" + self.expid
        self._allow_arrays = False
        self._allow_wrappers = True
        self.update_cmds()

        exp_id_path = os.path.join(BasicConfig.LOCAL_ROOT_DIR, self.expid)
        tmp_path = os.path.join(exp_id_path, "tmp")
        self._submit_script_path = os.path.join(
            tmp_path, BasicConfig.LOCAL_ASLOG_DIR, "submit_" + self.name + ".sh")
        # No need to write from API
        # self._submit_script_file = open(self._submit_script_path, 'w').close()

    def open_submit_script(self):
        # No need to write from API
        pass
        #self._submit_script_file = open(self._submit_script_path, 'w').close()
        #self._submit_script_file = open(self._submit_script_path, 'a')

    def get_submit_script(self):
        # No need to write from API
        pass
        # self._submit_script_file.close()
        # os.chmod(self._submit_script_path, 0o750)
        # return os.path.join(BasicConfig.LOCAL_ASLOG_DIR,os.path.basename(self._submit_script_path))

    def submit_Script(self):
        """
        Sends a SubmitfileScript, execute it  in the platform and retrieves the Jobs_ID of all jobs at once.

        :param job: job object
        :type job: autosubmit.job.job.Job
        :return: job id for  submitted jobs
        :rtype: list(int)
        """

        self.send_file(self.get_submit_script(), False)

        #cmd = '(cd '+self.get_files_path()+';'+' ./'+os.path.basename(self._submit_script_path)+')'

        cmd = os.path.join(self.get_files_path(),
                           os.path.basename(self._submit_script_path))
        if self.send_command(cmd):
            jobs_id = self.get_submitted_job_id(self.get_ssh_output())
            return jobs_id
        else:

            return None

    def update_cmds(self):
        """
        Updates commands for platforms
        """
        self.root_dir = os.path.join(
            self.scratch, self.project, self.user, self.expid)
        self.remote_log_dir = os.path.join(self.root_dir, "LOG_" + self.expid)
        self.cancel_cmd = "scancel"
        self._checkhost_cmd = "echo 1"
        self._submit_cmd = 'sbatch -D {1} {1}/'.format(
            self.host, self.remote_log_dir)
        self.put_cmd = "scp"
        self.get_cmd = "scp"
        self.mkdir_cmd = "mkdir -p " + self.remote_log_dir

    def get_checkhost_cmd(self):
        return self._checkhost_cmd

    def get_mkdir_cmd(self):
        return self.mkdir_cmd

    def get_remote_log_dir(self):
        return self.remote_log_dir

    def parse_job_output(self, output):
        return output.strip().split(' ')[0].strip()

    def parse_Alljobs_output(self, output, job_id):
        status = [x.split()[1] for x in output.splitlines()
                  if x.split()[0] == str(job_id)]
        if len(status) == 0:
            return status
        return status[0]

    def get_submitted_job_id(self, outputlines):
        if outputlines.find("failed") != -1:
            raise Exception(outputlines)
        jobs_id = []
        for output in outputlines.splitlines():
            jobs_id.append(int(output.split(' ')[3]))
        return jobs_id

    def jobs_in_queue(self):
        dom = parseString('')
        jobs_xml = dom.getElementsByTagName("JB_job_number")
        return [int(element.firstChild.nodeValue) for element in jobs_xml]

    def get_submit_cmd(self, job_script, job):
        self._submit_script_file.write(self._submit_cmd + job_script + "\n")

    def get_checkjob_cmd(self, job_id):
        return 'sacct -n -j {1} -o "State"'.format(self.host, job_id)

    def get_checkAlljobs_cmd(self, jobs_id):
        return "sacct -n -X -j  {1} -o 'jobid,State'".format(self.host, jobs_id)

    def get_queue_status_cmd(self, job_id):
        return 'squeue -j {0} -o %A,%R'.format(job_id)

    def parse_queue_reason(self, output, job_id):
        reason = [x.split(',')[1] for x in output.splitlines()
                  if x.split(',')[0] == str(job_id)]
        if len(reason) > 0:
            return reason[0]
        return reason
        # output = output.split('\n')
        # if len(output) > 1:
        #     return output[1]
        # else:
        #     return output

    @staticmethod
    def wrapper_header(filename, queue, project, wallclock, num_procs, dependency, directives):
        return """\
        #!/usr/bin/env python
        ###############################################################################
        #              {0}
        ###############################################################################
        #
        #SBATCH -J {0}
        {1}
        #SBATCH -A {2}
        #SBATCH --output={0}.out
        #SBATCH --error={0}.err
        #SBATCH -t {3}:00
        #SBATCH -n {4}
        {5}
        {6}
        #
        ###############################################################################
        """.format(filename, queue, project, wallclock, num_procs, dependency,
                   '\n'.ljust(13).join(str(s) for s in directives))

    @staticmethod
    def allocated_nodes():
        return """os.system("scontrol show hostnames $SLURM_JOB_NODELIST > node_list")"""



import os
import sys
import string
import time
import pickle
import textwrap
import traceback
import sqlite3
import copy
from datetime import datetime
from networkx import DiGraph

#DB_FILE_AS_TIMES = "/esarchive/autosubmit/as_times.db"


def get_structure(exp_id, structures_path):
    """
    Creates file of database and table of experiment structure if it does not exist. 
    Returns current structure as a Dictionary Job Name -> Children's Names

    :return: Map from job to children
    :rtype: Dictionary Key: String, Value: List(of String)
    """
    try:
        #pkl_path = os.path.join(exp_path, exp_id, "pkl")
        if os.path.exists(structures_path):
            db_structure_path = os.path.join(
                structures_path, "structure_" + exp_id + ".db")
            if not os.path.exists(db_structure_path):
                open(db_structure_path, "w")
            # print(db_structure_path)
            conn = create_connection(db_structure_path)
            create_table_query = textwrap.dedent(
                '''CREATE TABLE
            IF NOT EXISTS experiment_structure (
            e_from text NOT NULL,
            e_to text NOT NULL,
            UNIQUE(e_from,e_to)
            );''')
            create_table(conn, create_table_query)
            current_table = _get_exp_structure(db_structure_path)
            # print("Current table: ")
            # print(current_table)
            current_table_structure = dict()
            for item in current_table:
                _from, _to = item
                current_table_structure.setdefault(_from, []).append(_to)
                current_table_structure.setdefault(_to, [])
                # if _from not in current_table_structure.keys():
                #     current_table_structure[_from] = list()
                # if _to not in current_table_structure.keys():
                #     current_table_structure[_to] = list()
                # current_table_structure[_from].append(_to)
            if (len(current_table_structure.keys()) > 0):
                # print("Return structure")
                return current_table_structure
            else:
                return None
        else:
            # pkl folder not found
            raise Exception("structures folder not found " +
                            str(structures_path))
    except Exception as exp:
        print(traceback.format_exc())


def create_connection(db_file):
    """ 
    Create a database connection to the SQLite database specified by db_file.  
    :param db_file: database file name  
    :return: Connection object or None
    """
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except:
        return None


def create_table(conn, create_table_sql):
    """ create a table from the create_table_sql statement
    :param conn: Connection object
    :param create_table_sql: a CREATE TABLE statement
    :return:
    """
    try:
        c = conn.cursor()
        c.execute(create_table_sql)
    except Exception as e:
        print(e)


def _get_exp_structure(path):
    """
    Get all registers from experiment_status.\n
    :return: row content: exp_id, name, status, seconds_diff  
    :rtype: 4-tuple (int, str, str, int)
    """
    try:
        conn = create_connection(path)
        conn.text_factory = str
        cur = conn.cursor()
        cur.execute(
            "SELECT e_from, e_to FROM experiment_structure")
        rows = cur.fetchall()
        return rows
    except Exception as exp:
        print(traceback.format_exc())
        return dict()


def save_structure(graph, exp_id, structures_path):
    """
    Saves structure if path is valid
    """
    #pkl_path = os.path.join(exp_path, exp_id, "pkl")
    if os.path.exists(structures_path):
        db_structure_path = os.path.join(
            structures_path, "structure_" + exp_id + ".db")
        # with open(db_structure_path, "w"):
        conn = create_connection(db_structure_path)
        deleted = _delete_table_content(conn)
        if deleted == True:
            nodes_edges = {u for u, v in graph.edges()}
            nodes_edges.update({v for u, v in graph.edges()})
            independent_nodes = {u for u in graph.nodes() if u not in nodes_edges}
            data = {(u, v) for u, v in graph.edges()}
            data.update({(u, u) for u in independent_nodes})
            # save
            _create_edge(conn, data)
            #print("Created edge " + str(u) + str(v))
            conn.commit()
    else:
        # pkl folder not found
        raise Exception("pkl folder not found " + str(structures_path))


def _create_edge(conn, data):
    """
    Create edge
    """
    try:
        sql = ''' INSERT INTO experiment_structure(e_from, e_to) VALUES(?,?) '''
        cur = conn.cursor()
        cur.executemany(sql, data)
        # return cur.lastrowid
    except sqlite3.Error as e:
        print("Error on Insert : " + str(type(e).__name__))


def _delete_table_content(conn):
    """
    Deletes table content
    """
    try:
        sql = ''' DELETE FROM experiment_structure '''
        cur = conn.cursor()
        cur.execute(sql)
        conn.commit()
        return True
    except Exception as exp:
        # print(traceback.format_exc())
        print("Error on Delete _delete_table_content: {}".format(str(exp)))
        return False



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.


class FixedSizeList(list):
    """
        Customized list to retrieve a fixed sublist / slice of its elements.
        The list will be filled with as many items (default) as needed to
        get a list with the expected size.
    """

    def __init__(self, lst, default=None):
        super(FixedSizeList, self).__init__(lst)
        self._default = default

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [list(self)[i] if 0 <= i < len(self) else self._default for i in xrange(key.start, key.stop, key.step or 1)]
        return list(self)[key]

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))



#!/usr/bin/env python

# Copyright 2018 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import textwrap


class WrapperDirector:
    """
    Construct an object using the Builder interface.
    """

    def __init__(self):
        self._builder = None

    def construct(self, builder):
        self._builder = builder

        header = self._builder.build_header()
        job_thread = self._builder.build_job_thread()

        main = self._builder.build_main()

        # change to WrapperScript object
        wrapper_script = header + job_thread + main
        wrapper_script = wrapper_script.replace("_NEWLINE_", '\\n')

        return wrapper_script


class WrapperBuilder(object):

    def __init__(self, **kwargs):
        self.header_directive = kwargs['header_directive']
        self.job_scripts = kwargs['jobs_scripts']
        self.num_procs = kwargs['num_processors']
        self.expid = kwargs['expid']
        self.jobs_resources = kwargs.get('jobs_resources', dict())
        self.allocated_nodes = kwargs.get('allocated_nodes', '')
        self.machinefiles_name = ''
        self.machinefiles_indent = 0
        self.exit_thread = ''

    def build_header(self):
        return textwrap.dedent(self.header_directive) + self.build_imports()

    def build_imports(self):
        pass

    def build_job_thread(self):
        pass

    # hybrids
    def build_joblist_thread(self, **kwargs):
        pass

    # horizontal and hybrids
    def build_nodes_list(self):
        pass

    def build_machinefiles(self):
        pass

    def get_machinefile_function(self):
        machinefile_function = ""
        if 'MACHINEFILES' in self.jobs_resources and self.jobs_resources['MACHINEFILES']:
            machinefile_function = self.jobs_resources['MACHINEFILES']

            self.machinefiles_name = "jobname"

            if machinefile_function == 'COMPONENTS':
                return self.build_machinefiles_components()
            else:
                return self.build_machinefiles_standard()
        return machinefile_function

    def build_machinefiles_standard(self):
        pass

    def build_machinefiles_components(self):
        pass

    def build_machinefiles_components_alternate(self):
        pass

    def build_sequential_threads_launcher(self, **kwargs):
        pass

    def build_parallel_threads_launcher(self, **kwargs):
        pass

    # all should override -> abstract!
    def build_main(self):
        pass

    def _indent(self, text, amount, ch=' '):
        padding = amount * ch
        return ''.join(padding + line for line in text.splitlines(True))


class PythonWrapperBuilder(WrapperBuilder):

    def build_imports(self):
        return textwrap.dedent("""
        import os
        import sys
        from threading import Thread
        from commands import getstatusoutput
        from datetime import datetime
        from math import ceil
        from collections import OrderedDict
        import copy

        # Defining scripts to be run
        scripts = {0}
        """).format(str(self.job_scripts), '\n'.ljust(13))

    def build_job_thread(self):
        return textwrap.dedent("""
        class JobThread(Thread):
            def __init__ (self, template, id_run):
                Thread.__init__(self)
                self.template = template
                self.id_run = id_run

            def run(self):
                jobname = self.template.replace('.cmd', '')
                os.system("echo $(date +%s) > "+jobname+"_STAT")
                out = str(self.template) + "." + str(self.id_run) + ".out"
                err = str(self.template) + "." + str(self.id_run) + ".err"
                command = "bash " + str(self.template) + " " + str(self.id_run) + " " + os.getcwd()
                (self.status) = getstatusoutput(command + " > " + out + " 2> " + err)
        """).format('\n'.ljust(13))

    # hybrids
    def build_joblist_thread(self):
        pass

    # horizontal and hybrids
    def build_nodes_list(self):
        return self.get_nodes() + self.build_cores_list()

    def get_nodes(self):
        return textwrap.dedent("""
        # Getting the list of allocated nodes
        {0}
        os.system("mkdir -p machinefiles")

        with open('node_list', 'r') as file:
             all_nodes = file.read()

        all_nodes = all_nodes.split("_NEWLINE_")
        """).format(self.allocated_nodes, '\n'.ljust(13))

    def build_cores_list(self):
        return textwrap.dedent("""
        total_cores = {0}
        jobs_resources = {1}

        processors_per_node = int(jobs_resources['PROCESSORS_PER_NODE'])

        idx = 0
        all_cores = []
        while total_cores > 0:
            if processors_per_node > 0:
                processors_per_node -= 1
                total_cores -= 1
                all_cores.append(all_nodes[idx])
            else:
                idx += 1
                processors_per_node = int(jobs_resources['PROCESSORS_PER_NODE'])

        processors_per_node = int(jobs_resources['PROCESSORS_PER_NODE'])
        """).format(self.num_procs, str(self.jobs_resources), '\n'.ljust(13))

    def build_machinefiles(self):
        machinefile_function = self.get_machinefile_function()
        if machinefile_function:
            return self.get_machinefile_function() + self._indent(self.write_machinefiles(), self.machinefiles_indent)
        return ""

    def build_machinefiles_standard(self):
        return textwrap.dedent("""
        machines = str()

        cores = int(jobs_resources[section]['PROCESSORS'])
        tasks = int(jobs_resources[section]['TASKS'])
        nodes = int(ceil(int(cores)/float(tasks)))
        if tasks < processors_per_node:
            cores = tasks
       
        job_cores = cores
        while nodes > 0:
            while cores > 0:
                if len(all_cores) > 0:
                    node = all_cores.pop(0)
                    if node:
                        machines += node +"_NEWLINE_"
                        cores -= 1
            for rest in range(processors_per_node-tasks):
                if len(all_cores) > 0:
                    all_cores.pop(0)
            nodes -= 1
            if tasks < processors_per_node:
                cores = job_cores
        """).format('\n'.ljust(13))

    def _create_components_dict(self):
        return textwrap.dedent("""
        xio_procs = int(jobs_resources[section]['COMPONENTS']['XIO_NUMPROC'])
        rnf_procs = int(jobs_resources[section]['COMPONENTS']['RNF_NUMPROC'])
        ifs_procs = int(jobs_resources[section]['COMPONENTS']['IFS_NUMPROC'])
        nem_procs = int(jobs_resources[section]['COMPONENTS']['NEM_NUMPROC'])
        
        components = OrderedDict([
            ('XIO', xio_procs),
            ('RNF', rnf_procs),
            ('IFS', ifs_procs),
            ('NEM', nem_procs)
        ])
        
        jobs_resources[section]['COMPONENTS'] = components
        """).format('\n'.ljust(13))

    def build_machinefiles_components(self):
        return textwrap.dedent("""
        {0}
        
        machines = str()
        for component, cores in jobs_resources[section]['COMPONENTS'].items():        
            while cores > 0:
                if len(all_cores) > 0:
                    node = all_cores.pop(0)
                    if node:
                        machines += node +"_NEWLINE_"
                        cores -= 1
        """).format(self._create_components_dict(), '\n'.ljust(13))

    def write_machinefiles(self):
        return textwrap.dedent("""
        machines = "_NEWLINE_".join([s for s in machines.split("_NEWLINE_") if s])
        with open("machinefiles/machinefile_"+{0}, "w") as machinefile:
            machinefile.write(machines)
        """).format(self.machinefiles_name, '\n'.ljust(13))

    def build_sequential_threads_launcher(self, jobs_list, thread, footer=True):
        sequential_threads_launcher = textwrap.dedent("""
        for i in range(len({0})):
            current = {1}
            current.start()
            current.join()
        """).format(jobs_list, thread, '\n'.ljust(13))

        if footer:
            sequential_threads_launcher += self._indent(textwrap.dedent("""
            completed_filename = {0}[i].replace('.cmd', '_COMPLETED')
            completed_path = os.path.join(os.getcwd(), completed_filename)
            if os.path.exists(completed_path):
                print datetime.now(), "The job ", current.template," has been COMPLETED"
            else:
                print datetime.now(), "The job ", current.template," has FAILED"
                {1}
            """).format(jobs_list, self.exit_thread, '\n'.ljust(13)), 4)

        return sequential_threads_launcher

    def build_parallel_threads_launcher(self, jobs_list, thread, footer=True):
        parallel_threads_launcher = textwrap.dedent("""
        pid_list = []

        for i in range(len({0})):
            if type({0}[i]) != list:
                job = {0}[i]
                jobname = job.replace(".cmd", '')
                section = jobname.split('_')[-1]

            {2}

            current = {1}({0}[i], i)
            pid_list.append(current)
            current.start()

        # Waiting until all scripts finish
        for i in range(len(pid_list)):
            pid = pid_list[i]
            pid.join()
        """).format(jobs_list, thread, self._indent(self.build_machinefiles(), 8), '\n'.ljust(13))

        if footer:
            parallel_threads_launcher += self._indent(textwrap.dedent("""
            completed_filename = {0}[i].replace('.cmd', '_COMPLETED')
            completed_path = os.path.join(os.getcwd(), completed_filename)
            if os.path.exists(completed_path):
                print datetime.now(), "The job ", pid.template," has been COMPLETED"
            else:
                print datetime.now(), "The job ", pid.template," has FAILED"
                {1}
            """).format(jobs_list, self.exit_thread, '\n'.ljust(13)), 4)

        return parallel_threads_launcher

    # all should override -> abstract!
    def build_main(self):
        pass

    def dependency_directive(self):
        pass

    def queue_directive(self):
        pass

    def _indent(self, text, amount, ch=' '):
        padding = amount * ch
        return ''.join(padding + line for line in text.splitlines(True))


class PythonVerticalWrapperBuilder(PythonWrapperBuilder):

    def build_main(self):
        self.exit_thread = "os._exit(1)"
        return self.build_sequential_threads_launcher("scripts", "JobThread(scripts[i], i)")


class PythonHorizontalWrapperBuilder(PythonWrapperBuilder):

    def build_main(self):
        nodelist = self.build_nodes_list()
        threads_launcher = self.build_parallel_threads_launcher("scripts", "JobThread")
        return nodelist + threads_launcher


class PythonVerticalHorizontalWrapperBuilder(PythonWrapperBuilder):

    def build_joblist_thread(self):
        return textwrap.dedent("""
        class JobListThread(Thread):
            def __init__ (self, jobs_list, id_run):
                Thread.__init__(self)
                self.jobs_list = jobs_list
                self.id_run = id_run

            def run(self):
                {0}
        """).format(
            self._indent(self.build_sequential_threads_launcher("self.jobs_list", "JobThread(self.jobs_list[i], i)"),
                         8), '\n'.ljust(13))

    def build_main(self):
        self.exit_thread = "sys.exit()"
        joblist_thread = self.build_joblist_thread()
        nodes_list = self.build_nodes_list()
        threads_launcher = self.build_parallel_threads_launcher("scripts", "JobListThread", footer=False)
        return joblist_thread + nodes_list + threads_launcher


class PythonHorizontalVerticalWrapperBuilder(PythonWrapperBuilder):

    def build_joblist_thread(self):
        return textwrap.dedent("""
        class JobListThread(Thread):
            def __init__ (self, jobs_list, id_run, all_cores):
                Thread.__init__(self)
                self.jobs_list = jobs_list
                self.id_run = id_run
                self.all_cores = all_cores

            def run(self):
                all_cores = self.all_cores
                {0}
        """).format(
            self._indent(self.build_parallel_threads_launcher("self.jobs_list", "JobThread"), 8), '\n'.ljust(13))

    def build_main(self):
        nodes_list = self.build_nodes_list()
        self.exit_thread = "os._exit(1)"
        joblist_thread = self.build_joblist_thread()
        threads_launcher = self.build_sequential_threads_launcher("scripts", "JobListThread(scripts[i], i, "
                                                                             "copy.deepcopy(all_cores))", footer=False)
        return joblist_thread + nodes_list + threads_launcher


class BashWrapperBuilder(WrapperBuilder):

    def build_imports(self):
        return ""

    def build_main(self):
        return textwrap.dedent("""
        # Initializing variables
        scripts="{0}"
        i=0
        pids=""
        """).format(' '.join(str(s) for s in self.job_scripts), '\n'.ljust(13))

    def build_job_thread(self):
        return textwrap.dedent("""
        execute_script()
            {{
                out="$1.$2.out"
                err="$1.$2.err"
                bash $1 > $out 2> $err &
                pid=$!
            }}
        """).format('\n'.ljust(13))

    def build_sequential_threads_launcher(self):
        return textwrap.dedent("""
        for script in $scripts; do
            execute_script "$SCRATCH/{0}/LOG_{0}/$script" $i
            wait $pid
            if [ $? -eq 0 ]; then
                echo "The job $script has been COMPLETED"
            else
                echo "The job $script has FAILED"
                exit 1
            fi
            i=$((i+1))
        done
        """).format(self.expid, '\n'.ljust(13))

    def build_parallel_threads_launcher(self):
        return textwrap.dedent("""
        for script in $scripts; do
            execute_script "$SCRATCH/{0}/LOG_{0}/$script" $i
            pids+="$pid "
            i=$((i+1))
        done

        # Waiting until all scripts finish
        for pid in $pids; do
            wait $pid
            if [ $? -eq 0 ]; then
                echo "The job $pid has been COMPLETED"
            else
                echo "The job $pid has FAILED"
            fi
        done
        """).format(self.expid, '\n'.ljust(13))


class BashVerticalWrapperBuilder(BashWrapperBuilder):

    def build_main(self):
        return super(BashVerticalWrapperBuilder, self).build_main() + self.build_sequential_threads_launcher()


class BashHorizontalWrapperBuilder(BashWrapperBuilder):

    def build_main(self):
        return super(BashHorizontalWrapperBuilder, self).build_main() + self.build_parallel_threads_launcher()



from random import randrange
from unittest import TestCase

import os
from mock import Mock

from bscearth.utils.config_parser import ConfigParserFactory
from autosubmit.job.job import Job
from autosubmit.job.job_common import Status
from autosubmit.job.job_list import JobList
from autosubmit.job.job_common import Type
from autosubmit.job.job_list_persistence import JobListPersistenceDb


class TestJobList(TestCase):
    def setUp(self):
        self.experiment_id = 'random-id'
        self.job_list = JobList(self.experiment_id, FakeBasicConfig, ConfigParserFactory(),
                                JobListPersistenceDb('.', '.'))

        # creating jobs for self list
        self.completed_job = self._createDummyJobWithStatus(Status.COMPLETED)
        self.completed_job2 = self._createDummyJobWithStatus(Status.COMPLETED)
        self.completed_job3 = self._createDummyJobWithStatus(Status.COMPLETED)
        self.completed_job4 = self._createDummyJobWithStatus(Status.COMPLETED)

        self.submitted_job = self._createDummyJobWithStatus(Status.SUBMITTED)
        self.submitted_job2 = self._createDummyJobWithStatus(Status.SUBMITTED)
        self.submitted_job3 = self._createDummyJobWithStatus(Status.SUBMITTED)

        self.running_job = self._createDummyJobWithStatus(Status.RUNNING)
        self.running_job2 = self._createDummyJobWithStatus(Status.RUNNING)

        self.queuing_job = self._createDummyJobWithStatus(Status.QUEUING)

        self.failed_job = self._createDummyJobWithStatus(Status.FAILED)
        self.failed_job2 = self._createDummyJobWithStatus(Status.FAILED)
        self.failed_job3 = self._createDummyJobWithStatus(Status.FAILED)
        self.failed_job4 = self._createDummyJobWithStatus(Status.FAILED)

        self.ready_job = self._createDummyJobWithStatus(Status.READY)
        self.ready_job2 = self._createDummyJobWithStatus(Status.READY)
        self.ready_job3 = self._createDummyJobWithStatus(Status.READY)

        self.waiting_job = self._createDummyJobWithStatus(Status.WAITING)
        self.waiting_job2 = self._createDummyJobWithStatus(Status.WAITING)

        self.unknown_job = self._createDummyJobWithStatus(Status.UNKNOWN)

        self.job_list._job_list = [self.completed_job, self.completed_job2, self.completed_job3, self.completed_job4,
                                   self.submitted_job, self.submitted_job2, self.submitted_job3, self.running_job,
                                   self.running_job2, self.queuing_job, self.failed_job, self.failed_job2,
                                   self.failed_job3, self.failed_job4, self.ready_job, self.ready_job2,
                                   self.ready_job3, self.waiting_job, self.waiting_job2, self.unknown_job]

    def test_get_job_list_returns_the_right_list(self):
        job_list = self.job_list.get_job_list()
        self.assertEquals(self.job_list._job_list, job_list)

    def test_get_completed_returns_only_the_completed(self):
        completed = self.job_list.get_completed()

        self.assertEquals(4, len(completed))
        self.assertTrue(self.completed_job in completed)
        self.assertTrue(self.completed_job2 in completed)
        self.assertTrue(self.completed_job3 in completed)
        self.assertTrue(self.completed_job4 in completed)

    def test_get_submitted_returns_only_the_submitted(self):
        submitted = self.job_list.get_submitted()

        self.assertEquals(3, len(submitted))
        self.assertTrue(self.submitted_job in submitted)
        self.assertTrue(self.submitted_job2 in submitted)
        self.assertTrue(self.submitted_job3 in submitted)

    def test_get_running_returns_only_which_are_running(self):
        running = self.job_list.get_running()

        self.assertEquals(2, len(running))
        self.assertTrue(self.running_job in running)
        self.assertTrue(self.running_job2 in running)

    def test_get_running_returns_only_which_are_queuing(self):
        queuing = self.job_list.get_queuing()

        self.assertEquals(1, len(queuing))
        self.assertTrue(self.queuing_job in queuing)

    def test_get_failed_returns_only_the_failed(self):
        failed = self.job_list.get_failed()

        self.assertEquals(4, len(failed))
        self.assertTrue(self.failed_job in failed)
        self.assertTrue(self.failed_job2 in failed)
        self.assertTrue(self.failed_job3 in failed)
        self.assertTrue(self.failed_job4 in failed)

    def test_get_ready_returns_only_the_ready(self):
        ready = self.job_list.get_ready()

        self.assertEquals(3, len(ready))
        self.assertTrue(self.ready_job in ready)
        self.assertTrue(self.ready_job2 in ready)
        self.assertTrue(self.ready_job3 in ready)

    def test_get_waiting_returns_only_which_are_waiting(self):
        waiting = self.job_list.get_waiting()

        self.assertEquals(2, len(waiting))
        self.assertTrue(self.waiting_job in waiting)
        self.assertTrue(self.waiting_job2 in waiting)

    def test_get_unknown_returns_only_which_are_unknown(self):
        unknown = self.job_list.get_unknown()

        self.assertEquals(1, len(unknown))
        self.assertTrue(self.unknown_job in unknown)

    def test_get_in_queue_returns_only_which_are_queuing_submitted_and_running(self):
        in_queue = self.job_list.get_in_queue()

        self.assertEquals(7, len(in_queue))
        self.assertTrue(self.queuing_job in in_queue)
        self.assertTrue(self.running_job in in_queue)
        self.assertTrue(self.running_job2 in in_queue)
        self.assertTrue(self.submitted_job in in_queue)
        self.assertTrue(self.submitted_job2 in in_queue)
        self.assertTrue(self.submitted_job3 in in_queue)
        self.assertTrue(self.unknown_job in in_queue)

    def test_get_not_in_queue_returns_only_which_are_waiting_and_ready(self):
        not_in_queue = self.job_list.get_not_in_queue()

        self.assertEquals(5, len(not_in_queue))
        self.assertTrue(self.waiting_job in not_in_queue)
        self.assertTrue(self.waiting_job2 in not_in_queue)
        self.assertTrue(self.ready_job in not_in_queue)
        self.assertTrue(self.ready_job2 in not_in_queue)
        self.assertTrue(self.ready_job3 in not_in_queue)

    def test_get_finished_returns_only_which_are_completed_and_failed(self):
        finished = self.job_list.get_finished()

        self.assertEquals(8, len(finished))
        self.assertTrue(self.completed_job in finished)
        self.assertTrue(self.completed_job2 in finished)
        self.assertTrue(self.completed_job3 in finished)
        self.assertTrue(self.completed_job4 in finished)
        self.assertTrue(self.failed_job in finished)
        self.assertTrue(self.failed_job2 in finished)
        self.assertTrue(self.failed_job3 in finished)
        self.assertTrue(self.failed_job4 in finished)

    def test_get_active_returns_only_which_are_in_queue_ready_and_unknown(self):
        active = self.job_list.get_active()

        self.assertEquals(10, len(active))
        self.assertTrue(self.queuing_job in active)
        self.assertTrue(self.running_job in active)
        self.assertTrue(self.running_job2 in active)
        self.assertTrue(self.submitted_job in active)
        self.assertTrue(self.submitted_job2 in active)
        self.assertTrue(self.submitted_job3 in active)
        self.assertTrue(self.ready_job in active)
        self.assertTrue(self.ready_job2 in active)
        self.assertTrue(self.ready_job3 in active)
        self.assertTrue(self.unknown_job in active)

    def test_get_job_by_name_returns_the_expected_job(self):
        job = self.job_list.get_job_by_name(self.completed_job.name)

        self.assertEquals(self.completed_job, job)

    def test_sort_by_name_returns_the_list_of_jobs_well_sorted(self):
        sorted_by_name = self.job_list.sort_by_name()

        for i in xrange(len(sorted_by_name) - 1):
            self.assertTrue(sorted_by_name[i].name <= sorted_by_name[i + 1].name)

    def test_sort_by_id_returns_the_list_of_jobs_well_sorted(self):
        sorted_by_id = self.job_list.sort_by_id()

        for i in xrange(len(sorted_by_id) - 1):
            self.assertTrue(sorted_by_id[i].id <= sorted_by_id[i + 1].id)

    def test_sort_by_type_returns_the_list_of_jobs_well_sorted(self):
        sorted_by_type = self.job_list.sort_by_type()

        for i in xrange(len(sorted_by_type) - 1):
            self.assertTrue(sorted_by_type[i].type <= sorted_by_type[i + 1].type)

    def test_sort_by_status_returns_the_list_of_jobs_well_sorted(self):
        sorted_by_status = self.job_list.sort_by_status()

        for i in xrange(len(sorted_by_status) - 1):
            self.assertTrue(sorted_by_status[i].status <= sorted_by_status[i + 1].status)

    def test_that_create_method_makes_the_correct_calls(self):
        parser_mock = Mock()
        parser_mock.read = Mock()

        factory = ConfigParserFactory()
        factory.create_parser = Mock(return_value=parser_mock)

        job_list = JobList(self.experiment_id, FakeBasicConfig, factory, JobListPersistenceDb('.', '.'))
        job_list._create_jobs = Mock()
        job_list._add_dependencies = Mock()
        job_list.update_genealogy = Mock()
        job_list._job_list = [Job('random-name', 9999, Status.WAITING, 0),
                              Job('random-name2', 99999, Status.WAITING, 0)]
        date_list = ['fake-date1', 'fake-date2']
        member_list = ['fake-member1', 'fake-member2']
        num_chunks = 999
        chunk_list = range(1, num_chunks + 1)
        parameters = {'fake-key': 'fake-value',
                      'fake-key2': 'fake-value2'}
        graph_mock = Mock()
        job_list.graph = graph_mock
        # act
        job_list.generate(date_list, member_list, num_chunks, 1, parameters, 'H', 9999, Type.BASH, 'None')

        # assert
        self.assertEquals(job_list.parameters, parameters)
        self.assertEquals(job_list._date_list, date_list)
        self.assertEquals(job_list._member_list, member_list)
        self.assertEquals(job_list._chunk_list, range(1, num_chunks + 1))
        parser_mock.read.assert_called_once_with(os.path.join(FakeBasicConfig.LOCAL_ROOT_DIR, self.experiment_id,
                                                              'conf', "jobs_" + self.experiment_id + ".conf"))
        cj_args, cj_kwargs = job_list._create_jobs.call_args
        self.assertEquals(parser_mock, cj_args[1])
        self.assertEquals(0, cj_args[2])
        job_list._add_dependencies.assert_called_once_with(date_list, member_list, chunk_list, cj_args[0], parser_mock,
                                                           graph_mock)
        job_list.update_genealogy.assert_called_once_with(True, False)
        for job in job_list._job_list:
            self.assertEquals(parameters, job.parameters)

    def test_that_create_job_method_calls_dic_jobs_method_with_increasing_priority(self):
        # arrange
        dic_mock = Mock()
        dic_mock.read_section = Mock()

        parser_mock = Mock()
        parser_mock.sections = Mock(return_value=['fake-section-1',
                                                  'fake-section-2'])
        # act
        JobList._create_jobs(dic_mock, parser_mock, 0, Type.BASH)

        # arrange
        dic_mock.read_section.assert_any_call('fake-section-1', 0, Type.BASH, dict())
        dic_mock.read_section.assert_any_call('fake-section-2', 1, Type.BASH, dict())

    def _createDummyJobWithStatus(self, status):
        job_name = str(randrange(999999, 999999999))
        job_id = randrange(1, 999)
        job = Job(job_name, job_id, status, 0)
        job.type = randrange(0, 2)
        return job


class FakeBasicConfig:
    def __init__(self):
        pass

    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''



import os
import sys
import string
import time
import pickle
import textwrap
import traceback
import sqlite3
import copy
from datetime import datetime
from collections import OrderedDict

DB_FILE_AS_TIMES = "/esarchive/autosubmit/as_times.db"
DB_FILES_STATUS = "/esarchive/autosubmit/as_metadata/test/status.db"
# PATH_DB_DATA = "/esarchive/autosubmit/as_metadata/data/"


def prepare_status_db():
    """
    Creates table of experiment status if it does not exist.  
    :return: Map from experiment name to (Id of experiment, Status, Seconds)  
    :rtype: Dictionary Key: String, Value: Integer, String, Integer
    """
    conn = create_connection(DB_FILE_AS_TIMES)
    create_table_query = textwrap.dedent(
        '''CREATE TABLE
    IF NOT EXISTS experiment_status (
    exp_id integer PRIMARY KEY,
    name text NOT NULL,
    status text NOT NULL,
    seconds_diff integer NOT NULL,
    modified text NOT NULL,
    FOREIGN KEY (exp_id) REFERENCES experiment (id)
    );''')
    drop_table_query = ''' DROP TABLE experiment_status '''
    # create_table(conn, drop_table_query)
    create_table(conn, create_table_query)
    current_table = _get_exp_status()
    current_table_expid = dict()
    # print(current_table)
    # print(type(current_table))
    for item in current_table:
        exp_id, expid, status, seconds = item
        current_table_expid[expid] = (exp_id, status, seconds)
    return current_table_expid


def prepare_completed_times_db():
    """
    Creates two tables: 
    "experiment_times" that stores general information about experiments' jobs completed updates
    "job_times" that stores information about completed time for jobs in the database as_times.db
    Then retrieves all the experiments data from "experiment_times".  
    :return: Dictionary that maps experiment name to experiment data.  
    :rtype: Dictionary: Key String, Value 6-tuple (exp_id, name, created, modified, total_jobs, completed_jobs)
    """
    conn = create_connection(DB_FILE_AS_TIMES)
    create_table_header_query = textwrap.dedent(
        '''CREATE TABLE
    IF NOT EXISTS experiment_times (
    exp_id integer PRIMARY KEY,
    name text NOT NULL,
    created int NOT NULL,    
    modified int NOT NULL,
    total_jobs int NOT NULL,
    completed_jobs int NOT NULL,
    FOREIGN KEY (exp_id) REFERENCES experiment (id)
    );''')

    create_table_detail_query = textwrap.dedent(
        '''CREATE TABLE
    IF NOT EXISTS job_times (
    detail_id integer PRIMARY KEY AUTOINCREMENT,
    exp_id integer NOT NULL,    
    job_name text NOT NULL,    
    created integer NOT NULL,    
    modified integer NOT NULL,
    submit_time int NOT NULL,
    start_time int NOT NULL,
    finish_time int NOT NULL,
    status text NOT NULL,
    FOREIGN KEY (exp_id) REFERENCES experiment (id)
    );''')

    drop_table_header_query = ''' DROP TABLE experiment_times '''
    drop_table_details_query = ''' DROP TABLE job_times '''
    # create_table(conn, drop_table_details_query)
    # create_table(conn, drop_table_header_query)
    # return
    create_table(conn, create_table_header_query)
    create_table(conn, create_table_detail_query)
    current_table = _get_exp_times()
    current_table_expid = dict()

    for item in current_table:
        try:
            exp_id, name, created, modified, total_jobs, completed_jobs = item
            current_table_expid[name] = (exp_id, int(created), int(
                modified), int(total_jobs), int(completed_jobs))
        except Exception as ex:
            print(traceback.format_exc())
            print(item)
            print(str(name) + " ~ " + str(created) + "\t" + str(modified))

    return current_table_expid

# STATUS ARCHIVE


def insert_archive_status(status, alatency, abandwidth, clatency, cbandwidth, rtime):

    try:
        conn = create_connection(DB_FILES_STATUS)
        sql = ''' INSERT INTO archive_status(status, avg_latency, avg_bandwidth, current_latency, current_bandwidth, response_time, modified ) VALUES(?,?,?,?,?,?,?)'''
        # print(row_content)
        cur = conn.cursor()
        cur.execute(sql, (int(status), alatency, abandwidth, clatency,
                          cbandwidth, rtime, datetime.today().strftime('%Y-%m-%d-%H:%M:%S')))
        # print(cur)
        conn.commit()
        return cur.lastrowid
    except Exception as exp:
        print(traceback.format_exc())
        print("Error on Insert : " + str(exp))


def get_last_read_archive_status():
    """

    :return: status, avg. latency, avg. bandwidth, current latency, current bandwidth, response time, date
    :rtype: 7-tuple
    """
    try:
        conn = create_connection(DB_FILES_STATUS)
        sql = "SELECT status, avg_latency, avg_bandwidth, current_latency, current_bandwidth, response_time, modified FROM archive_status order by rowid DESC LIMIT 1"
        cur = conn.cursor()
        cur.execute(sql)
        rows = cur.fetchall()
        status, alatency, abandwidth, clatency, cbandwidth, rtime, date = rows[0]
        return (status, alatency, abandwidth, clatency, cbandwidth, rtime, date)
        # print(rows)
    except Exception as exp:
        print(traceback.format_exc())
        print("Error on Get Last : " + str(exp))
        return (False, None, None, None, None, None, None)

# INSERTIONS


def insert_experiment_times_header(expid, timest, total_jobs, completed_jobs, debug=False, conn_ecearth=None):
    """
    Inserts into experiment times header. Requires ecearth.db connection.    
    :param expid: Experiment name  
    :type expid: String  
    :param timest: timestamp of the pkl last modified date  
    :type timest: Integer  
    :param total_jobs: Total number of jobs  
    :type total_jobs: Integer  
    :param completed_jobs: Number of completed jobs  
    :type completed_jobs: Integer  
    :param debug: Flag (testing purposes)  
    :type debug: Boolean  
    :param conn_ecearth: ecearth.db connection  
    :type conn: sqlite3 connection
    :return: exp_id of the experiment (not the experiment name)  
    :rtype: Integer
    """
    try:
        current_id = _get_id_db(conn_ecearth, expid)
        if (current_id):
            if (debug == True):
                print("INSERTING INTO EXPERIMENT_TIMES " + str(current_id) + "\t" + str(expid) +
                      "\t" + str(timest) + "\t" + str(total_jobs) + "\t" + str(completed_jobs))
                return current_id
            row_content = (current_id, expid, int(timest),
                           int(timest), total_jobs, completed_jobs)
            result = _create_exp_times(row_content)
            return current_id
        else:
            return -1
    except sqlite3.Error as e:
        print("Error on Insert : " + str(type(e).__name__))
        print(current_id)


def _create_exp_times(row_content):
    """
    Create experiment times
    :param conn:
    :param details:
    :return:
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        sql = ''' INSERT INTO experiment_times(exp_id, name, created, modified, total_jobs, completed_jobs) VALUES(?,?,?,?,?,?) '''
        # print(row_content)
        cur = conn.cursor()
        cur.execute(sql, row_content)
        # print(cur)
        conn.commit()
        return cur.lastrowid
    except sqlite3.Error as e:
        print("Error on Insert : " + str(type(e).__name__))
        print(row_content)


def create_job_times(conn, exp_id, job_name, timest, submit_time, start_time, finish_time, status, debug=False):
    """
    Create job_times detail register.  
    :param conn: connection to as_times.db  
    :type conn: sqlite3 connection  
    :param exp_id: Id of experiment (not name)  
    :type exp_id: Integer  
    :param job_name: Name of job  
    :type job_name: String  
    :param timest: Timestamp of the last modified date of the pkl file  
    :type timest: Integer  
    :param submit_time: Timestamp of the submit time of the job  
    :type submit_time: Integer  
    :param start_time: Timestamp of the start time of the job  
    :type start_time: Integer  
    :param finish_time: Timestamp of the finish time of the job  
    :type finish_time: Integer  
    :param status: Status of the job  
    :type status: String  
    :param debug: Flag for testing purposes  
    :type debug: Boolean  
    :return: Id of detail  
    :rtype: Integer
    """

    try:
        if (debug == True):
            print("INSERTING INTO JOB_TIMES " + str(exp_id) + " ~ " + str(job_name) + " ~ " + str(timest) +
                  "\t" + str(submit_time) + "\t" + str(start_time) + "\t" + str(finish_time) + "\t" + str(status))
            return
        sql = ''' INSERT INTO job_times(exp_id, job_name, created, modified, submit_time, start_time, finish_time, status) VALUES(?,?,?,?,?,?,?,?) '''
        cur = conn.cursor()
        cur.execute(sql, (exp_id, job_name, timest, timest,
                          submit_time, start_time, finish_time, status))
        # Commit outside the loop
        conn.commit()
        return cur.lastrowid
    except sqlite3.Error as e:
        print(traceback.format_exc())
        print("Error on Insert : " + str(type(e).__name__))
        print((exp_id, job_name, timest, timest,
               submit_time, start_time, finish_time, status))


def create_many_job_times(conn, list_job):
    try:
        sql = ''' INSERT INTO job_times(exp_id, job_name, created, modified, submit_time, start_time, finish_time, status) VALUES(?,?,?,?,?,?,?,?) '''
        # print(row_content)
        cur = conn.cursor()
        cur.executemany(sql, list_job)
        # print(cur)
        # Commit outside the loop
        conn.commit()
    except sqlite3.Error as e:
        print(traceback.format_exc())
        print("Error on Insert : " + str(type(e).__name__))
        # print((exp_id, job_name, timest, timest,
        #        submit_time, start_time, finish_time, status))


def _insert_into_ecearth_details(conn, exp_id, user, created, model, branch, hpc):
    """
    Inserts into the details table of ecearth.db  
    :return: Id  
    :rtype: int
    """
    if conn:
        try:
            sql = ''' INSERT INTO details(exp_id, user, created, model, branch, hpc) VALUES(?,?,?,?,?,?) '''
            cur = conn.cursor()
            cur.execute(sql, (exp_id, user, created, model, branch, hpc))
            conn.commit()            
            return cur.lastrowid
        except Exception as exp:
            print(exp)
    return False


def insert_experiment_status(conn, expid, seconds_diff):
    """
    Inserts into table of status if it does not exist and updates status. Requires conn to ecearth.db.\n
    :param conn: ecearth.db connection  
    :type conn: sqlite3 connection object  
    :param expid: Experiment name  
    :type expid: str  
    :param seconds_diff: time limit in seconds  
    :type seconds_diff: int  
    :return: Experiment id  
    :rtype: int
    """
    # print("Looking {}".format(expid))
    current_id = _get_id_db(conn, expid)
    conn = create_connection(DB_FILE_AS_TIMES)
    if (current_id):
        try:
            # print("Inserting {} {}".format(current_id, expid))
            row_content = (current_id, expid, "RUNNING", int(
                seconds_diff), datetime.today().strftime('%Y-%m-%d-%H:%M:%S'))
            result = _create_exp_status(row_content)
        except sqlite3.Error as exp:
            print("error on insert_experiment_status {0} {1}".format(
                current_id, expid))
            print(traceback.format_exc())
            return current_id
        # print(row_content)

    return current_id


def create_connection(db_file):
    """ 
    Create a database connection to the SQLite database specified by db_file.  
    :param db_file: database file name  
    :return: Connection object or None
    """
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except Exception as exp:
        print(exp)


def create_table(conn, create_table_sql):
    """ create a table from the create_table_sql statement
    :param conn: Connection object
    :param create_table_sql: a CREATE TABLE statement
    :return:
    """
    try:
        c = conn.cursor()
        c.execute(create_table_sql)
    except Exception as e:
        print(e)


# SELECTS

def get_completed_times_detail():
    """
    Gets all details in job times as a Dictionary of dictionaries
    """
    # experiments = get_exps(conn)
    try:
        # conn = create_connection(DB_FILE_AS_TIMES)
        all_detail = _get_all_job_times()
        exps = dict()
        for item in all_detail:
            detail_id, exp_id, job_name, created, modified, submit_time, start_time, finish_time, status = item
            if exp_id not in exps.keys():
                exps[exp_id] = dict()
            exps[exp_id][job_name] = (
                submit_time, start_time, finish_time, status, detail_id)
        return exps
    except Exception as ex:
        print(traceback.format_exc())


def get_times_detail(exp_id):
    """
    Gets the current detail of the experiment from job_times  
    :param exp_id: Id of the experiment  
    :type exp_id: Integer  
    :return: Dictionary Key: Job Name, Values: 5-tuple (submit time, start time, finish time, status, detail id)  
    :rtype: dict
    """
    # conn = create_connection(DB_FILE_AS_TIMES)
    try:
        current_table_detail = dict()
        current_table = _get_job_times(exp_id)
        if current_table is None:
            return None
        for item in current_table:
            detail_id, exp_id, job_name, created, modified, submit_time, start_time, finish_time, status = item
            current_table_detail[job_name] = (
                submit_time, start_time, finish_time, status, detail_id)
        return current_table_detail
    except Exception as exp:
        print(traceback.format_exc())
        return None


def get_times_detail_by_expid(conn, expid):
    """
    Gets the detail of times of the experiment by expid (name of experiment).\n
    :param conn: ecearth.db connection  
    :rtype conn: sqlite3 connection  
    :param expid: Experiment name  
    :type expid: str  
    :return: Dictionary Key: Job Name, Values: 5-tuple (submit time, start time, finish time, status, detail id)  
    :rtype: dict
    """
    # conn = create_connection(DB_FILE_AS_TIMES)
    exp_id = _get_id_db(conn, expid)
    if (exp_id):
        return get_times_detail(exp_id)
    else:
        return None


def get_experiment_status():
    """
    Gets table experiment_status as dictionary
    conn is expected to reference as_times.db
    """
    # conn = create_connection(DB_FILE_AS_TIMES)
    experiment_status = dict()
    current_table = _get_exp_status()
    for item in current_table:
        exp_id, name, status, seconds_diff = item
        experiment_status[name] = status
    return experiment_status


def get_currently_running_experiments():
    """
    Gets the list of currently running experiments ordered by total_jobs  
    Connects to AS_TIMES  
    :return: map of expid -> total_jobs  
    :rtype: OrderedDict() str -> int
    """
    experiment_running = OrderedDict()
    current_running = _get_exp_only_active()
    if current_running:
        for item in current_running:
            name, status, total_jobs = item
            experiment_running[name] = total_jobs
    return experiment_running


def get_specific_experiment_status(expid):
    """
    Gets the current status from database.\n
    :param expid: Experiment name  
    :type expid: str  
    :return: name of experiment and status  
    :rtype: 2-tuple (name, status)
    """
    exp_id, name, status, seconds_diff = _get_specific_exp_status(expid)
    print("{} {} {} {}".format(exp_id, name, status, seconds_diff))
    return (name, status)


def get_experiment_times():
    """
    Gets table experiment_times as dictionary
    conn is expected to reference as_times.db
    """
    # conn = create_connection(DB_FILE_AS_TIMES)
    experiment_times = dict()
    current_table = _get_exp_times()
    for item in current_table:
        exp_id, name, created, modified, total_jobs, completed_jobs = item
        experiment_times[name] = (total_jobs, completed_jobs, modified)
        # if extended == True:
        #     experiment_times[name] = (total_jobs, completed_jobs, created, modified)
    return experiment_times


def get_experiment_times_by_expid(expid):
    """[summary] 
    :return: exp_id, total number of jobs, number of completed jobs 
    :rtype 3-tuple: (int, int, int) 
    """
    current_row = _get_exp_times_by_expid(expid)
    if current_row:
        exp_id, name, created, modified, total_jobs, completed_jobs = current_row
        return (exp_id, total_jobs, completed_jobs)
    return None


def get_experiment_times_group():
    """
    Gets table experiment_times as dictionary id -> name
    conn is expected to reference as_times.db
    """
    experiment_times = dict()
    current_table = _get_exp_times()
    for item in current_table:
        exp_id, name, created, modified, total_jobs, completed_jobs = item
        if name not in experiment_times.keys():
            experiment_times[name] = list()
        experiment_times[name].append(exp_id)
        # if extended == True:
        #     experiment_times[name] = (total_jobs, completed_jobs, created, modified)
    return experiment_times


def get_exps(conn):
    """
    Get exp ids
    :param conn: ecearth.db connection
    :param expid:
    :return: 
    """
    # SQLite always return a unicode object, but we can change this
    # behaviour with the next sentence
    conn.text_factory = str
    cur = conn.cursor()
    cur.execute(
        "SELECT name FROM experiment WHERE autosubmit_version IS NOT NULL")
    rows = cur.fetchall()
    return rows


def _get_exps_complete(conn):
    """
    Get exp all row information
    :param conn: ecearth.db connection
    :param expid:
    :return: 
    """
    # SQLite always return a unicode object, but we can change this
    # behaviour with the next sentence
    conn.text_factory = str
    cur = conn.cursor()
    cur.execute(
        "SELECT * FROM experiment WHERE autosubmit_version IS NOT NULL")
    rows = cur.fetchall()
    return rows


def get_exps_base(conn):
    """
    Get exp name and id from experiment table in ecearth.db
    :param conn: ecearth.db connection
    :param expid:
    :return: 
    """
    result = dict()
    conn.text_factory = str
    cur = conn.cursor()
    cur.execute(
        "SELECT name, id FROM experiment WHERE autosubmit_version IS NOT NULL")
    rows = cur.fetchall()
    #print("len " + str(len(rows)))
    if (rows):
        for row in rows:
            _name, _id = row
            result[_name] = _id
    else:
        return None
    return result


def get_exps_detailed_complete(conn):
    """
    Get information from details table
    :param conn: ecearth.db connection
    :param expid:
    :return: Dictionary exp_id -> (user, created, model, branch, hpc)
    """
    all_details = _get_exps_detailed_complete(conn)
    result = dict()
    if (all_details):
        for item in all_details:
            exp_id, user, created, model, branch, hpc = item
            result[exp_id] = (user, created, model, branch, hpc)
    return result


def get_latest_completed_jobs(seconds=300):
    """
    Get latest completed jobs
    """
    result = dict()
    latest_completed_detail = _get_latest_completed_jobs(seconds)
    if (latest_completed_detail):
        for item in latest_completed_detail:
            detail_id, exp_id, job_name, created, modified, submit_time, start_time, finish_time, status = item
            result[job_name] = (detail_id, submit_time,
                                start_time, finish_time, status)
    return result


def _get_exps_detailed_complete(conn):
    """
    Get information from details table
    :param conn: ecearth.db connection
    :param expid:
    :return:
    """
    conn.text_factory = str
    cur = conn.cursor()
    cur.execute(
        "SELECT * FROM details")
    rows = cur.fetchall()
    return rows


def _get_exp_times():
    """
    Get all experiments from table experiment_times.\n
    :return: Row content (exp_id, name, created, modified, total_jobs, completed_jobs)  
    :rtype: 6-tuple (int, str, str, str, int, int)
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        conn.text_factory = str
        cur = conn.cursor()
        cur.execute(
            "SELECT exp_id, name, created, modified, total_jobs, completed_jobs FROM experiment_times")
        rows = cur.fetchall()
        return rows
    except Exception as exp:
        print(traceback.format_exc())
        return list()


def _get_exp_times_by_expid(expid):
    """
    Returns data from experiment_time table by expid. 

    :param expid: expid/name
    :type expid: str
    :return: Row content (exp_id, name, created, modified, total_jobs, completed_jobs)  
    :rtype: 6-tuple (int, str, str, str, int, int)
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        conn.text_factory = str
        cur = conn.cursor()
        cur.execute(
            "SELECT exp_id, name, created, modified, total_jobs, completed_jobs FROM experiment_times WHERE name=?", (expid,))
        rows = cur.fetchall()
        return rows[0] if len(rows) > 0 else None
    except Exception as exp:
        print(traceback.format_exc())
        return None


def _get_job_times(exp_id):
    """
    Get exp job times detail for a given expid from job_times.  
    :param exp_id: Experiment id (not name)  
    :type exp_id: Integer  
    :return: Detail content  
    :rtype: list of 9-tuple: (detail_id, exp_id, job_name, created, modified, submit_time, start_time, finish_time, status)
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        cur.execute("SELECT detail_id, exp_id, job_name, created, modified, submit_time, start_time, finish_time, status FROM job_times WHERE exp_id=?", (exp_id,))
        rows = cur.fetchall()
        return rows
    except Exception as ex:
        print(traceback.format_exc())
        print(str(exp_id))
        print(str(ex))
        return None


def _get_latest_completed_jobs(seconds=300):
    """
    get latest completed jobs, defaults at 300 seconds, 5 minutes
    """
    current_ts = time.time()
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        cur.execute("SELECT detail_id, exp_id, job_name, created, modified, submit_time, start_time, finish_time, status FROM job_times WHERE (? - modified)<=? AND status='COMPLETED'", (current_ts, seconds))
        rows = cur.fetchall()
        return rows
    except Exception as ex:
        print(traceback.format_exc())
        print(str(seconds))
        print(str(ex))
        return None


def _get_all_job_times():
    """
    Get content from job_times detail for all experiments.\n
    :return: (detail_id, exp_id, job_name, created, modified, submit, start, finish_time, status)  
    :rtype: 9-tuple (int, int, str, int, int, int, int, int, str)
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        cur.execute(
            "SELECT detail_id, exp_id, job_name, created, modified, submit_time, start_time, finish_time, status FROM job_times")
        rows = cur.fetchall()
        return rows
    except Exception as exp:
        print("Error while trying to _get_all_job_times")
        print(traceback.format_exc())
        return None


def _delete_from_job_times_detail(detail_id):
    """
    Deletes a job that is no longer included in the experiment. 

    :param detail_id: id of row  
    :type detail_id: int
    """
    try:
        #print("About to delete: " + str(detail_id))
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        #print("Cursor defined")
        cur.execute(
            "DELETE FROM job_times WHERE detail_id=?", (detail_id,))
        # print(cur.rowcount)
        cur.close()
        conn.commit()
        conn.close()

        # No reliable way to get any feedback from cursor at this point, so let's just return 1
        return True
    except Exception as exp:
        print("Error while trying to delete " +
              str(detail_id) + " from job_times.")
        print(traceback.format_exc())
        return None


def _delete_many_from_job_times_detail(detail_list):
    try:
        #print("About to delete: " + str(detail_id))
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        #print("Cursor defined")
        cur.executemany(
            "DELETE FROM job_times WHERE detail_id=?", detail_list)
        # print(cur.rowcount)
        cur.close()
        conn.commit()
        conn.close()
        # No reliable way to get any feedback from cursor at this point, so let's just return 1
        return True
    except Exception as exp:
        # print("Error while trying to delete " +
        #       str(detail_id) + " from job_times.")
        print(traceback.format_exc())
        return None


def delete_experiment_data(exp_id):
    """
    Deletes experiment data from experiment_times and job_times in as_times.db

    :param exp_id: Id of experiment 
    :type exp_id: int
    """
    deleted_e = _delete_from_experiment_times(exp_id)
    deleted_j = _delete_from_job_times(exp_id)
    print("Main exp " + str(deleted_e) + "\t" + str(deleted_j))
    if deleted_e and deleted_j:
        return True
    return None


def _delete_from_experiment_times(exp_id):
    """
    Deletes an experiment from experiment_times in as_times.db

    :param exp_id: Id of experiment 
    :type exp_id: int
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        cur.execute("DELETE FROM experiment_times WHERE exp_id=?", (exp_id,))
        conn.commit()
        conn.close()
        return True
    except Exception as exp:
        print("Error while trying to delete " +
              str(exp_id) + " from experiment_times.")
        print(traceback.format_exc())
        return None


def _delete_from_job_times(exp_id):
    """
    Deletes an experiment from job_times in as_times.db

    :param exp_id: Id of experiment 
    :type exp_id: int
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        cur.execute("DELETE FROM job_times WHERE exp_id=?", (exp_id,))
        conn.commit()
        conn.close()
        return True
    except Exception as exp:
        print("Error while trying to delete " +
              str(exp_id) + " from job_times.")
        print(traceback.format_exc())
        return None


def _get_exp_status():
    """
    Get all registers from experiment_status.\n
    :return: row content: exp_id, name, status, seconds_diff  
    :rtype: 4-tuple (int, str, str, int)
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        conn.text_factory = str
        cur = conn.cursor()
        cur.execute(
            "SELECT exp_id, name, status, seconds_diff FROM experiment_status")
        rows = cur.fetchall()
        return rows
    except Exception as exp:
        print(traceback.format_exc())
        return dict()


def _get_exp_only_active():
    """
    Get all registers of experiments ACTIVE
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        conn.text_factory = str
        cur = conn.cursor()
        cur.execute("select name, status, total_jobs from currently_running")
        rows = cur.fetchall()
        return rows
    except Exception as exp:
        print(traceback.format_exc())
        print(exp)
        return None


def _get_specific_exp_status(expid):
    """
    Get all registers from experiment_status.\n
    :return: row content: exp_id, name, status, seconds_diff  
    :rtype: 4-tuple (int, str, str, int)
    """
    try:
        # print("Honk")
        conn = create_connection(DB_FILE_AS_TIMES)
        cur = conn.cursor()
        cur.execute(
            "SELECT exp_id, name, status, seconds_diff FROM experiment_status WHERE name=?", (expid,))
        row = cur.fetchone()
        if row == None:
            return (0, expid, "NOT RUNNING", 0)
        # print(row)
        return row
    except Exception as exp:
        print(traceback.format_exc())
        return (0, expid, "NOT RUNNING", 0)


def _get_exp_status_running():
    """
    Get RUNNING experiments from from experiment_status.\n
    :return: row content: exp_id, name, status, seconds_diff  
    :rtype: 4-tuple (int, str, str, int)
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        conn.text_factory = str
        cur = conn.cursor()
        cur.execute(
            "SELECT exp_id, name, status, seconds_diff FROM experiment_status WHERE status='RUNNING'")
        rows = cur.fetchall()
        return rows
    except sqlite3.Error as e:
        print("Error on _get_exp_status_running")


def _create_exp_status(row_content):
    """
    Create experiment status
    :param conn:
    :param details:
    :return:
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        sql = ''' INSERT INTO experiment_status(exp_id, name, status, seconds_diff, modified) VALUES(?,?,?,?,?) '''
        # print(row_content)
        cur = conn.cursor()
        cur.execute(sql, row_content)
        # print(cur)
        conn.commit()
        return cur.lastrowid
    except sqlite3.Error as e:
        print(row_content)
        print("Error on Insert : " + str(type(e).__name__))


def _get_id_db(conn, expid):
    """
    Get exp_id of the experiment (different than the experiment name).  
    :param conn: ecearth.db connection  
    :type conn: sqlite3 connection  
    :param expid: Experiment name  
    :type expid: String  
    :return: Id of the experiment  
    :rtype: Integer or None
    """
    try:
        cur = conn.cursor()
        cur.execute("SELECT id FROM experiment WHERE name=?", (expid,))
        row = cur.fetchone()
        return row[0]
    except Exception as exp:
        print("Couldn't get exp_id for {0}".format(expid))
        return None


# UPDATES


def update_exp_status(expid, status, seconds_diff):
    """
    Update existing experiment_status.  
    :param expid: Experiment name  
    :type expid: String  
    :param status: Experiment status  
    :type status: String  
    :param seconds_diff: Indicator of how long it has been active since the last time it was checked  
    :type seconds_diff: Integer  
    :return: Id of register  
    :rtype: Integer
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        sql = ''' UPDATE experiment_status SET status = ?, seconds_diff = ?, modified = ? WHERE name = ? '''
        cur = conn.cursor()
        cur.execute(sql, (status, seconds_diff,
                          datetime.today().strftime('%Y-%m-%d-%H:%M:%S'), expid))
        conn.commit()
        return cur.lastrowid
    except sqlite3.Error as e:
        print("Error while trying to update " +
              str(expid) + " in experiment_status.")
        print(traceback.format_exc())
        print("Error on Update: " + str(type(e).__name__))


def _update_ecearth_details(conn, exp_id, user, created, model, branch, hpc):
    """
    Updates ecearth.db table details.
    """
    if conn:
        try:
            sql = ''' UPDATE details SET user=?, created=?, model=?, branch=?, hpc=? where exp_id=? '''
            cur = conn.cursor()
            cur.execute(sql, (user, created, model, branch, hpc, exp_id))
            conn.commit()
            return True
        except Exception as exp:
            print(exp)
            return False
    return False


def update_experiment_times(exp_id, modified, completed_jobs, total_jobs, debug=False):
    """
    Update existing experiment times header
    """
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        if (debug == True):
            print("UPDATE experiment_times " + str(exp_id) + "\t" +
                  str(completed_jobs) + "\t" + str(total_jobs))
            return
        sql = ''' UPDATE experiment_times SET modified = ?, completed_jobs = ?, total_jobs = ? WHERE exp_id = ? '''
        cur = conn.cursor()
        cur.execute(sql, (int(modified), completed_jobs, total_jobs, exp_id))
        conn.commit()
        # print("Updated header")
        return exp_id
    except sqlite3.Error as e:
        print("Error while trying to update " +
              str(exp_id) + " in experiment_times.")
        print(traceback.format_exc())
        print("Error on Update: " + str(type(e).__name__))


def update_experiment_times_only_modified(exp_id, modified):
    try:
        conn = create_connection(DB_FILE_AS_TIMES)
        sql = ''' UPDATE experiment_times SET modified = ? WHERE exp_id = ? '''
        cur = conn.cursor()
        cur.execute(sql, (int(modified), exp_id))
        conn.commit()
        # print("Updated header")
        return exp_id
    except sqlite3.Error as e:
        print("Error while trying to update " +
              str(exp_id) + " in experiment_times.")
        print(traceback.format_exc())
        print("Error on Update: " + str(type(e).__name__))


def update_job_times(conn, detail_id, modified, submit_time, start_time, finish_time, status, debug=False, no_modify_time=False):
    """
    Update single experiment job detail \n
    :param conn: Connection to as_times.db. \n
    :type conn: sqlite3 connection. \n
    :param detail_id: detail_id in as_times.job_times. \n
    :type detail_id: Integer. \n
    :param modified: Timestamp of current date of pkl \n
    :type modified: Integer. \n
    """
    try:
        #conn = create_connection(DB_FILE_AS_TIMES)
        # if submit_time > start_time or start_time > finish_time:
        #     print("[Warning] Update " + str(detail_id) + "\t" + str(submit_time) + "\t" +
        #           str(start_time) + "\t" + str(finish_time) + "\t" + str(status))
        #     return None
        # else:
        #     print("Update " + str(detail_id) + "\t" + str(submit_time) + "\t" +
        #           str(start_time) + "\t" + str(finish_time) + "\t" + str(status))
        if (debug == True):
            print("UPDATING JOB TIMES " + str(detail_id) + " ~ " + str(modified) + "\t" +
                  str(submit_time) + "\t" + str(start_time) + "\t" + str(finish_time) + "\t" + str(status))
            return
        if no_modify_time == False:
            sql = ''' UPDATE job_times SET modified = ?, submit_time = ?, start_time = ?, finish_time = ?, status = ? WHERE detail_id = ? '''
            cur = conn.cursor()
            cur.execute(sql, (modified, submit_time, start_time,
                              finish_time, status, detail_id))
            # Commit outside the loop
            conn.commit()
        else:
            sql = ''' UPDATE job_times SET submit_time = ?, start_time = ?, finish_time = ?, status = ? WHERE detail_id = ? '''
            cur = conn.cursor()
            cur.execute(sql, (submit_time, start_time,
                              finish_time, status, detail_id))
            # Commit outside the loop
            conn.commit()
        # print(str(detail_id) + " UPDATED to " + status)
    except sqlite3.Error as e:
        print("Error while trying to update " +
              str(detail_id) + " in job_times.")
        print(traceback.format_exc())
        print("Error on Update: " + str(type(e).__name__))


def update_many_job_times(conn, list_jobs):
    try:
        sql = ''' UPDATE job_times SET modified = ?, submit_time = ?, start_time = ?, finish_time = ?, status = ? WHERE detail_id = ? '''
        cur = conn.cursor()
        cur.executemany(sql, list_jobs)
        # Commit outside the loop
        conn.commit()
    except sqlite3.Error as e:
        print("Error while trying to update many in update_many_job_times.")
        print(traceback.format_exc())
        print("Error on Update: " + str(type(e).__name__))



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import os

from autosubmitAPIwu.platforms.paramiko_platform import ParamikoPlatform
from autosubmitAPIwu.platforms.headers.lsf_header import LsfHeader
from autosubmitAPIwu.platforms.wrappers.wrapper_factory import LSFWrapperFactory


class LsfPlatform(ParamikoPlatform):
    """
    Class to manage jobs to host using LSF scheduler

    :param expid: experiment's identifier
    :type expid: str
    """

    def __init__(self, expid, name, config):
        ParamikoPlatform.__init__(self, expid, name, config)
        self._header = LsfHeader()
        self._wrapper = LSFWrapperFactory(self)
        self.job_status = dict()
        self.job_status['COMPLETED'] = ['DONE']
        self.job_status['RUNNING'] = ['RUN']
        self.job_status['QUEUING'] = ['PEND', 'FW_PEND']
        self.job_status['FAILED'] = ['SSUSP', 'USUSP', 'EXIT']
        self._allow_arrays = True
        self._allow_wrappers = True
        # No need to write from API
        # self.update_cmds()

    def update_cmds(self):
        """
        Updates commands for platforms
        """
        self.root_dir = os.path.join(
            self.scratch, self.project, self.user, self.expid)
        self.remote_log_dir = os.path.join(self.root_dir, "LOG_" + self.expid)
        self.cancel_cmd = "bkill"
        self._checkjob_cmd = "bjobs "
        self._checkhost_cmd = "echo 1"
        self._submit_cmd = "bsub -cwd " + self.remote_log_dir + \
            " < " + self.remote_log_dir + "/"
        self.put_cmd = "scp"
        self.get_cmd = "scp"
        self.mkdir_cmd = "mkdir -p " + self.remote_log_dir

    def get_checkhost_cmd(self):
        return self._checkhost_cmd

    def get_mkdir_cmd(self):
        return self.mkdir_cmd

    def get_remote_log_dir(self):
        return self.remote_log_dir

    def parse_job_output(self, output):
        job_state = output.split('\n')
        if len(job_state) > 1:
            job_state = job_state[1].split()
            if len(job_state) > 2:
                return job_state[2]
        # If we can not process output, assuming completed. Then we will look for completed files and status will
        # change to failed if COMPLETED file is not present.
        return 'DONE'

    def get_submitted_job_id(self, output):
        return output.split('<')[1].split('>')[0]

    def jobs_in_queue(self):
        return zip(*[line.split() for line in ''.split('\n')])[0][1:]

    def get_checkjob_cmd(self, job_id):
        return self._checkjob_cmd + str(job_id)

    def get_submit_cmd(self, job_script, job):
        return self._submit_cmd + job_script

    @staticmethod
    def wrapper_header(filename, queue, project, wallclock, num_procs, dependency, directives):
        return """\
        #!/usr/bin/env python
        ###############################################################################
        #              {0}
        ###############################################################################
        #
        #BSUB -J {0}
        #BSUB -q {1}
        #BSUB -P {2}
        #BSUB -oo {0}.out
        #BSUB -eo {0}.err
        #BSUB -W {3}
        #BSUB -n {4}
        {5}
        {6}
        #
        ###############################################################################
        """.format(filename, queue, project, wallclock, num_procs, dependency,
                   '\n'.ljust(13).join(str(s) for s in directives))



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import json 

class Tree(object):

    def __init__(self, list_job):
        self.root = list()        
        self._listjob = list_job
        self.visited = list()
        for job in self._listjob:                       
            if job.has_parents() == False:
                self.root.append(job)
            
    def traverse_format(self):
        """
        Start recursive traversal and construction of tree
        """        
        result = list()
        for start in self.root:            
            self.visited.append(start.name)
            level = 0
            result.append({'title' : start.title, 'refKey': start.name, 'data': 'Empty', 'expanded': True , 'children' : self.recursive_add(start, level)})
        return result
        
    def recursive_add(self, job, level):
        res_data = list()
        level = level + 1
        for start in job._children:            
            if start.name not in self.visited:                
                self.visited.append(start.name)
                res_data.append({'title' : start.title, 'refKey': start.name, 'data': 'Empty', 'expanded': True if level <= 100 else False, 'children': self.recursive_add(start,level)})
            else:
                res_data.append({'title' : start.title, 'refKey': start.name, 'data': 'Empty', 'children': []})
        return res_data


        


# class Node(object):
#     def __init__(self, identifier, data = None):
#         self._identifier = identifier
#         self.data = data
#         self.children = {}